% :- type file      == atom.
% :- type row(T)    == list(T).
% :- type matrix(T) == list(row(T)).
% :- type dir       ---> n                % north
%                   ;    w                % west
%                   ;    s                % south
%		    ;    e.               % east
% :- type field     == int.
% :- type domino    ---> field-field.      

:- module(domino_frame, [
	domino_in/2,    % domino_in(file::in, matrix(field)::out)
	domino_out/3,	% domino_out(file::in, matrix(field)::in,
		 	%       matrix(dir)::in)
	solve/2,	% solve(file::in, file::in)
	time/2		% time(file::in, file::in)
			]
	 ).

:- use_module(library(lists), [append/3, nth/3, max_list/2]).
:- use_module(library(ordsets), [ord_subtract/3, list_to_ord_set/2]).

%-------------- Bejaratok ------------------------------------

% domino_in(+File, -Table):
% A File-bol beolvashato egy feladat, melyet a Table matrix ir le.
domino_in(File, Table) :-
	open_file(File, read, Stream),
	read_in(Stream, Characters),
	close(Stream),
	parse_data(Characters, Table).

% domino_out(+File, +Table, +Solution):
% A Table feladat Solution megoldasat kiirja a File allomanyba.
domino_out(File, Table, Solution) :-
	open_file(File, write, Stream),
	current_output(OOut),
	set_output(Stream),
	call_cleanup(pretty_print(Table, Solution),
		     (set_output(OOut),close(Stream))).

% solve(+FileIn, +FileOut):
% A FileIn allomanybol beolvasott feladat minden megoldasat
% kiirja a FileOut allomanyba.
solve(FileIn, FileOut) :-
	domino_in(FileIn, Table),
	ensure_loaded(user:domino),
	open_file(FileOut, write, Stream),
	current_output(OOut),
	set_output(Stream),
	call_cleanup(print_solutions(Table),
		     (set_output(OOut),close(Stream))).

% print_solutions(+Table):
% A Table feladat minden megoldasat kiirja az aktualis kimenetre.
print_solutions(Table) :-
	user:domino(Table, Solution),
	check_coverage(Table, Solution), 
	pretty_print(Table, Solution),
	fail.
print_solutions(_).

% time(+FileIn, +FileOut):
% Mint solve/2, de a futashoz szukseges idot is kiirja.
time(FileIn, FileOut) :-
	statistics(runtime, [T0|_]),
	solve(FileIn, FileOut),
	statistics(runtime, [T1|_]),
	T is T1 - T0,
	format('Table: ~a, time: ~3d sec~n', [FileIn,T]).

%-------------- Allomanyok kezelese --------------------------

% open_file(+File, +Mode, -Stream):
% A File allomanyt megnyitja a Stream folyam formajaban Mode modban.
open_file(user, M, S) :-
	!, standard_stream(M, S).
open_file(File, Mode, Stream) :-
	open(File, Mode, Stream).

% standard_stream(?Mode, ?Stream):
% A Mode modu szabvanyos folyam a Stream.
standard_stream(read, user_input).
standard_stream(write, user_output).

%-------------- Beolvasas ------------------------------------

% read_in(+S, -L):
% S folyambol beolvashato karakterek listaja L.
read_in(S, L) :-
	get_code(S, C),
	read_in(C, S, L).

% read_in(+C, +S, -L):
% Ha C az S folyambol elozoleg beolvasott karakter, akkor L az S-bol
% beolvashato maradek karakterek listaja, C-t is beleertve.  Ha C = -1,
% akkor L = [].
read_in(C, _, L) :-
	C =:= -1, !, L = [].
read_in(C, S, [C|L]) :-
	get_code(S, C1),
	read_in(C1, S, L).

% parse_data(+Chars, -Table):
% Chars fuzerbol kielemezheto egy feladat, amit a Table matrix ir le.
parse_data(Chars, Table) :-
	int_lists(Table, Chars, Rest),
	correct_syntax(Rest),
	correct_matrix(Table),
	correct_field_counts(Table).

%-------------- Ellenorzesek ---------------------------------

% correct_syntax(+Rest):
% A Rest fuzer a bemenet helyes maradeka, azaz ures.
correct_syntax(Rest) :- 
	Rest = [_|_], !, 
	chars_of_type(mid_line, L, Rest, _),
	warning('non-digit character in line: "~s"', [L]),
	fail.
correct_syntax(_).

% correct_matrix(+M):
% M valoban egy matrix, azaz minden sora egyforma hosszu.
correct_matrix([Row1|Rows]) :-
	length(Row1, Len1),
	nth(I, Rows, Row),
	length(Row, Len),
	Len =\= Len1, !,
	I1 is I+1,
	warning(
	  'non-rectangular: line 1 length: ~d, line ~d length: ~d',
	  [Len1, I1, Len]),
	fail.
correct_matrix(_).

% correct_field_counts(+Table):
% A Table tablazatban egy teljes dominokeszlet szamai vannak.
correct_field_counts(Table) :-
	matrix_elements_max(Table, All, N),
	between(0, N, I),
	count(I, All, Count),
	Count =\= N+2, !,
	N2 is N+2, 
	warning(
	  'incorrect table: count of ~d fields = ~d, should be ~d (max = ~d)',
	   [I, Count, N2, N]),
	fail.
correct_field_counts(_).

% check_coverage(+Table, +Solution):
% Mindig sikerul, de figyelmeztet, ha a Solution megoldas a
% Table tablazattal nem tartalmazza egy teljes dominokeszlet elemeit.
check_coverage(Table, Solution) :-
	missing_dominos(Table, Solution, Doms),
	Doms \== [], !, 
	warning('not all dominos used, missing: ~w', [Doms]).
check_coverage(_, _).

% missing_dominos(+Table, +Solution, -Doms):
% A Solution megoldas a Table tablazattal nem tartalmazza a Doms
% listaban szereplo dominokat.
missing_dominos(Table, Solution, Doms) :-
	matrix_elements_max(Table, _, N),
	findall(D, domino_on_table(Table, Solution, D), TableDoms),
	findall(D1-D2, (between(0, N, D2), between(0, D2, D1)), AllDoms),
	list_to_ord_set(TableDoms, TDoms),
	list_to_ord_set(AllDoms, ADoms),
	ord_subtract(ADoms, TDoms, Doms).

% domino_on_table(Table, Solution, Dom):
% A Dom domino szerepel a Table tablazaton a Solution megoldassal.
domino_on_table(Table, Solution, Dom) :-
	pair_matrices(Solution, Table, ST),
	append(_, [Row|Rows], ST),
	(   append(_, [S1-D1,S2-D2|_], Row),
	    S1 == w, S2 == e	% behelyettesitetlen Si eseten nem sikerul
	;   Rows = [NextRow|_],
	    par_member(S1-D1, Row, S2-D2, NextRow),
	    S1 == n, S2 == s
	),
	canonical_pair(D1, D2, Dom).

% canonical_pair(D1, D2, Dom):
% A D1 es D2 szamokat tartalmazo domino kanonikus formaja Dom.
canonical_pair(D1, D2, Dom) :-
	D1 =< D2, !, Dom = D1-D2.
canonical_pair(D1, D2, D2-D1).

%-------------- DCG elemzes ----------------------------------

% int_lists(-LL) -->
% Kielemezheto az LL egeszek listainak listaja.
int_lists([IntList|IntLists]) -->
	chars_of_type(layout, _),
	int_list(IntList), !, 
	int_lists(IntLists).
int_lists([]) --> [].

% int_list(-IntList) -->
% Kielemezheto az IntList egeszek listaja, ujsorral lezarva.  
% A bemenet nem kezdodhet nem lathato karakterrel.
int_list([Int|IntList]) -->
	int_number(Int), !,
	chars_of_type(horiz_sp, _),
	int_list(IntList).
int_list([]) -->
	char_of_type(vert_sp, _).

% int_number(-Int) -->
% Kielemezheto egy (tizes szamrendszerbeli) Int nem-negativ egesz szam.
int_number(Int) -->
	chars_of_type(digit, Ds),
	{Ds = [_|_], number_codes(Int, Ds)}.

% layout_char(Dir, C):
% C egy Dir (horiz_sp|vert_sp) iranyu nem lathato karakter.
layout_char(horiz_sp, 0' ).
layout_char(horiz_sp, 0'\t).
layout_char(vert_sp,  0'\n).
layout_char(vert_sp,  0'\r).
layout_char(vert_sp,  0'\f).
layout_char(vert_sp,  0'\v).

% Kielemezheto egy Type tipusu C karakter.
char_of_type(Type, C) -->
	[C], {char_type(Type, C)}.

% chars_of_type(Type, Cs) -->
% Kielemezheto Type tipusu karakterek egy Cs listaja.	
chars_of_type(Type, [C|Cs]) -->
	char_of_type(Type, C), !,
	chars_of_type(Type, Cs).
chars_of_type(_, []) --> [].

% char_type(Type, C) :
% A C karakter Type tipusu.
char_type(layout, C) :-
	layout_char(_, C).
char_type(digit, D) :-
	D >= 0'0, D =< 0'9.
char_type(mid_line, C) :-
	\+ layout_char(vert_sp, C).
char_type(Dir, C) :-
	layout_char(Dir, C).

%-------------- Kiiratas -------------------------------------

% pretty_print(+Table, +Solution):
% kiirja a Table feladat  Solution megoldasat.
pretty_print(Table, Solution) :-
	Table = [FRow|_],
	length(FRow, NumCols),
	matrix_elements_max(Table, _, N),
	Pad is integer(log(10, max(N,1)))+1,
	IWidth is (Pad+3)*NumCols+1,
	format('~*c~n', [IWidth, 0'-]),
	pretty_print_rows(Table, Solution, Pad),
	format('~*c~n~n', [IWidth, 0'-]).

% pretty_print_rows(+Rs, +Ss, +Pad):
% Az Rs feladat- es Ss megoldassorokat kiirja, a szamokat egy Pad
% szelessegu mezoben jobbra igazitva.
pretty_print_rows([R|Rs], [S|Ss], Pad) :-
	put_code(0'|), pretty_print_row(R, S, Pad),
	(   Ss = [NextS|_] -> 
	    put_code(0'|), pretty_print_underline(S, NextS, Pad),
	    pretty_print_rows(Rs, Ss, Pad)
	;   /* Ss = [],  */ true
	).

% pretty_print_row(+R, +S, +Pad):
% Az R feladat- es az S megoldassort (a szamokat egy min. Pad
% szelessegu mezoben jobbra igazitva) kiirja.
pretty_print_row([], [], _) :-
	nl.
pretty_print_row([V|Vs], Ds0, Pad) :-
	right_char(Ds0, RightC),
	format(' ~|~t~d~*+ ~c', [V,Pad,RightC]),
	Ds0 = [_|Ds],
	pretty_print_row(Vs, Ds, Pad).

% pretty_print_underline(Ds, NextDs, Pad):
% Kiirja a Ds iranyokhoz tartozo alahuzas-sort, feltetelezve
% hogy az alatta levo iranyok NextDs, es egy szam hossza Pad.
pretty_print_underline([], [], _) :-
	nl.
pretty_print_underline(Ds0, NDs0, Pad) :-
	Ds0 = [D|Ds], NDs0 = [ND|NDs], 
	bottom_char(D, ND, BottomC),
	corner_char(Ds0, NDs0, CornerC),
	N is Pad+2,
	format('~*c~c', [N, BottomC,CornerC]),
	pretty_print_underline(Ds, NDs, Pad).

% Az alabbi harom eljaras megprobal ertelmes hatarolo karaktereket
% kiadni akkor is, ha a megoldas-matrix nem teljesen kitoltott.

% right_char(?Ds, ?Right):
% A Ds iranylista elso eleme jobboldalan kiirando karakter a Right.
right_char(Ds, Right) :-
	\+ Ds = [w,e|_], !, Right = 0'|.
right_char(_, 0' ).

% bottom_char(?D, ?ND, ?Bottom):
% A D es ND iranyok koze illo also karakter a Bottom.
bottom_char(D, ND, Bottom) :-
	(\+ D = n ; \+ ND = s), !, Bottom = 0'-.
bottom_char(_D, _ND, 0' ).

% corner_char(?Ds, ?NDs, ?Corner):
% A Ds, es az alatta levo NDs iranylistakhoz tartozo sarok-karakter a
% Corner.
corner_char(Ds, NDs, Corner) :-
	\+ Ds = [w,e|_], \+ NDs = [w,e|_], !, Corner = 0'|.
corner_char([D|_], [ND|_], Corner) :-
	( D == w ; ND == w ), !, Corner = 0'-.
corner_char(_, _, 0' ).

%-------------- Seged-eljarasok ------------------------------

% between(+N, +M, -X):
% Az X egesz szam az [N, M] intervallum eleme.
between(N, M, N) :-
	N =< M.
between(N, M, X) :-
	N < M, N1 is N+1,
	between(N1, M, X).

% append_lists(+Ls, ?L):
% Az Ls listak osszefuzesevel kapott lista L.
append_lists([], []).
append_lists([H|T], C) :-
	append(H, R, C),
	append_lists(T, R).

% count(+Elem, +List, ?Cnt):
% Elem a List lista elemekent Cnt alkalommal fordul elo.
count(Elem, List, Cnt) :-
	count(List, Elem, 0, Cnt).

% count(+Elem, +List, +Cnt0, ?Cnt):
% Elem a List lista elemekent Cnt-Cnt0 alkalommal fordul elo.
count([], _, C, C).
count([E|L], E, C0, C) :-
	!, C1 is C0+1,
	count(L, E, C1, C).
count([_|L], E, C0, C) :-
	count(L, E, C0, C).

% matrix_elements_max(+LLs, ?All, ?Max):
% Az LLs matrix osszes elemenek listaja All, ezek maximuma Max.
matrix_elements_max(LLs, All, Max) :-
	append_lists(LLs, All),
	max_list(All, Max).

% pair_matrices(+M1, +M2, ?PM):
% M1, M2 es PM azonos alaku matrixok, ahol PMij = M1ij-M2ij,
% minden ervenyes i sor- es j oszlop-indexre.
pair_matrices([], [], []).
pair_matrices([R1|R1s], [R2|R2s], [PR|PRs]) :-
	pair_list(R1, R2, PR),
	pair_matrices(R1s, R2s, PRs).

% pair_list(+L1, +L2, ?PL):
% L1, L2 es PL azonos hosszu listak, ahol PLi = L1i-L2i,
% minden ervenyes i indexre.
pair_list([], [], []).
pair_list([E1|E1s], [E2|E2s], [E1-E2|Ps]) :-
	pair_list(E1s, E2s, Ps).

% par_member(?X, +XL, ?Y, +YL):
% X es Y rendre az XL es YL listak azonos indexu eleme.
par_member(X1, [X1|_], X2, [X2|_]).
par_member(X1, [_|L1], X2, [_|L2]) :-
	par_member(X1, L1, X2, L2).

% warning(+Txt, +Args):
% Kiirja a Txt szoveggel es Args argumentumlistaval
% megadott figyelmezteto szoveget.
warning(Txt, Args) :-
	print_message(warning, format(Txt, Args)).

%-------------- ISO kompatibilitas ---------------------------

% Az alabbi cel az ISO Prolog szabvanynak meg nem megfelelo, de
% DEC10 kompatiblis Prolog rendszerekben valo futtatashoz szukseges.

:- (   predicate_property(get_code(_,_), _) -> true
   ;   assert((get_code(S, C) :- get0(S, C))),
       assert((put_code(C) :- put(C))),
       assert((number_codes(N, C) :- number_chars(N, C)))
   ).

